// Copyright 2002 Adam Megacz, see the COPYING file for licensing [GPL]
package org.xwt;

import java.io.*;
import java.net.*;
import java.text.*;
import java.util.*;
import org.xwt.util.*;
import org.mozilla.javascript.*;
import org.bouncycastle.util.encoders.Base64;

/** Singleton class that provides all functionality in the xwt.* namespace */
public final class XWT extends JSObject {

    public static final XWT singleton = new XWT();

    /** each key is a string representing a filename which the user has already given XWT permission to write to */
    private static Hashtable safeFiles = new Hashtable();

    public String getClassName() { return "XWT"; }
    private XWT() { setSeal(true); }

    public Object get(String name, Scriptable start) {
        if (name == null) return null;
        else if (name.equals("maxdim")) return new Integer(Short.MAX_VALUE);
        else if (name.equals("parseFloat")) return JSObject.defaultObjects.get("parseFloat", null);
        else if (name.equals("parseInt")) return JSObject.defaultObjects.get("parseInt", null);
        else if (name.equals("alt")) return Surface.alt ? Boolean.TRUE : Boolean.FALSE;
        else if (name.equals("control")) return Surface.control ? Boolean.TRUE : Boolean.FALSE;
        else if (name.equals("shift")) return Surface.shift ? Boolean.TRUE : Boolean.FALSE;
        else if (name.equals("date")) return date;
        else if (name.equals("listfonts")) return listfonts;
        else if (name.equals("regexp")) return regexp;
        else if (name.equals("sleep")) return sleep;
        else if (name.equals("yield")) return yield;
        else if (name.equals("newBrowserWindow")) return newBrowserWindow;
        else if (name.equals("textwidth")) return textwidth;
        else if (name.equals("textheight")) return textheight;
        else if (name.equals("newBox")) return newBox;
        else if (name.equals("soap")) return soap;
        else if (name.equals("xmlrpc")) return xmlrpc;
        else if (name.equals("origin")) return Main.origin;
        else if (name.equals("clipboard")) return Platform.getClipBoard();
        else if (name.equals("altKeyName")) return Platform.altKeyName();
        else if (name.equals("screenWidth")) return new Integer(Platform.getScreenWidth());
        else if (name.equals("screenHeight")) return new Integer(Platform.getScreenHeight());
        else if (name.equals("static")) return Static.getStatic("");
        else if (name.equals("theme")) return theme;
        else if (name.equals("openFile")) return openFile;
        else if (name.equals("saveFile")) return saveFile;
        else if (name.equals("saveFileAs")) return saveFileAs;
        else if (name.equals("utfEncode")) return utfEncode;
        else if (name.equals("fileSeparator")) return File.separator;
        else if (name.equals("homeDir")) return System.getProperty("user.home");
        else if (name.equals("tempDir")) return System.getProperty("java.io.tempdir");
        else if (name.equals("recursivePrintObject")) return recursivePrintObject;
        else if (name.equals("parseHTML")) return parseHTML;
        else if (name.equals("button")) {
            if (Surface.button1 && !Surface.button2 && !Surface.button3) return new Integer(1);
            else if (!Surface.button1 && Surface.button2 && !Surface.button3) return new Integer(1);
            else if (!Surface.button1 && !Surface.button2 && Surface.button3) return new Integer(1);
            else return new Integer(0);
        }
        else if (name.equals("println")) return println;
        else if (name.equals("math")) return org.xwt.util.JSObject.defaultObjects.get("Math", null);
        else if (name.equals("loadArchive")) return loadArchive;
        else if (name.equals("prefetchImage")) return prefetchImage;
        else if (name.equals("prefs")) return prefs;
        else if (name.equals("encodeURI")) return JSObject.defaultObjects.get("encodeURI", null);
        else if (name.equals("encodeURIComponent")) return JSObject.defaultObjects.get("encodeURIComponent", null);
        else if (name.equals("decodeURI")) return JSObject.defaultObjects.get("decodeURI", null);
        else if (name.equals("decodeURIComponent")) return JSObject.defaultObjects.get("decodeURIComponent", null);
        else return super.get(name, start);
    }

    public void put(String name, Scriptable start, Object value) {
        if (name == null) return;
        else if (name.equals("thread") && value != null && value instanceof Function) ThreadMessage.newthread((Function)value);
        else if (name.equals("clipboard")) Platform.setClipBoard(value.toString());

        // FIXME: undocumented, possibly insecure
        else if (name.equals("proxyAuthorization")) {
            Proxy.Authorization.authorization = value.toString();
            Proxy.Authorization.waitingForUser.release();
        }

        else super.put(name, start, value);
    }


    // Prefs Object //////////////////////////////////////////////////////////////////////////

    static Scriptable prefsRPC = new XMLRPC("http://megacz:mypassword@localhost/RPC2", "prefs");

    private static final JSObject prefs = new JSObject(false, true) {
            public Object get(String name, Scriptable start) {
                if (name.equals("get")) return prefsGet;
                else if (name.equals("list")) return prefsList;
                else if (name.equals("put")) return prefsPut;
                else return null;
            }
        };

    private static final JSObject.JSFunction prefsGet = new JSObject.JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                if (args.length != 1 || args[0] == null) return null;
                try {
                    return ((Function)prefsRPC.get("get", null)).call(cx, null, null, new Object[] { args[0] });
                } catch (JavaScriptException jse) {
                    Object val = jse.getValue();
                    if (val instanceof JSObject && new Integer(1).equals(((JSObject)val).get("faultCode")))
                        return null;
                    throw jse;
                } catch (Exception e) {
                    // FIXME
                    throw new JavaScriptException(e.toString());
                }
            }
        };

    private static final JSObject.JSFunction prefsPut = new JSObject.JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                if (args.length < 2 || args[0] == null) return null;
                try {
                    return ((Function)prefsRPC.get("put", null)).call(cx, null, null, new Object[] { args[0].toString(), args[1] });
                } catch (JavaScriptException jse) {
                    Object val = jse.getValue();
                    if (val instanceof JSObject && new Integer(1).equals(((JSObject)val).get("faultCode")))
                        return null;
                    throw jse;
                } catch (Exception e) {
                    // FIXME
                    throw new JavaScriptException(e.toString());
                }
            }
        };

    private static final JSObject.JSFunction prefsList = new JSObject.JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                if (args.length < 2 || args[0] == null) return null;
                try {
                    return ((Function)prefsRPC.get("list", null)).call(cx, null, null, new Object[] { args[0].toString(), args[1] });
                } catch (JavaScriptException jse) {
                    Object val = jse.getValue();
                    if (val instanceof JSObject && new Integer(1).equals(((JSObject)val).get("faultCode")))
                        return null;
                    throw jse;
                } catch (Exception e) {
                    // FIXME
                    throw new JavaScriptException(e.toString());
                }
            }
        };

    private static final JSObject.JSFunction prefsInvoke = new JSObject.JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                if (args.length < 2 || args[0] == null) return null;
                try {
                    return ((Function)prefs.get("invoke")).call(cx, null, null, new Object[] { args[0].toString(), args[1], "megacz", "mypassword" });
                } catch (JavaScriptException jse) {
                    Object val = jse.getValue();
                    if (val instanceof JSObject && new Integer(1).equals(((JSObject)val).get("faultCode")))
                        return null;
                    throw jse;
                } catch (Exception e) {
                    // FIXME
                    throw new JavaScriptException(e.toString());
                }
            }
        };



    // JSFunction Instances ///////////////////////////////////////////////////////////////////

    private static final JSObject.JSFunction newBrowserWindow = new JSObject.JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                if (args.length != 1 || args[0] == null) return null;
                Platform.newBrowserWindow(args[0].toString());
                return null;
            }
        };

    public static final JSObject.JSFunction yield = new JSObject.JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                sleep.call(cx, null, null, null);
                return null;
            }
        };

    private static final JSObject.JSFunction println = new JSObject.JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                if (args.length == 1)
                    if (Log.on) {
                        String source = cx.interpreterSourceFile;
                        if (source.endsWith("._")) source = source.substring(0, source.length() - 2);
                        Log.log(source, args[0] == null ? "null" : args[0].toString());
                    }
                return null;
            }
        };

    private static final JSObject.JSFunction date = new JSObject.JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                try { return Context.enter().newObject(org.xwt.util.JSObject.defaultObjects, "Date", args);
                } catch (Exception e) {
                    if (Log.on) Log.log(this, "Exception in Context.newObject() -- this should never happen");
                    if (Log.on) Log.log(this, e);
                    return null;
                }
            }
        };

    private static final JSObject.JSFunction regexp = new JSObject.JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                try { return Context.enter().newObject(org.xwt.util.JSObject.defaultObjects, "RegExp", args);
                } catch (Exception e) {
                    if (Log.on) Log.log(this, "Exception in Context.newObject() -- this should never happen");
                    if (Log.on) Log.log(this, e);
                    return null;
                }
            }
        };

    private static final JSObject.JSFunction listfonts = new JSObject.JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                return Context.enter().newArray(org.xwt.util.JSObject.defaultObjects,  Platform.listFonts());
            }
        };

    private static final JSObject.JSFunction theme = new JSObject.JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                if (args.length != 2) return null;
                if (args[0] == null || args[1] == null) return null;

                for(int i=1; i<args.length; i++) {
                    if (args[i] instanceof String) {
                        String from = (String)args[0];
                        String to = (String)args[i];
                        if (Log.on) Log.log(this, "retheming from " + from + " to " + to);
                        Resources.mapFrom.addElement(from);
                        Resources.mapTo.addElement(to);
                    }
                }

                Function callback = args[args.length - 1] instanceof Function ? (Function)args[args.length - 1] : null;
                Template.retheme(callback);
                return null;
            }
        };

    private static final JSObject.JSFunction xmlrpc = new JSObject.JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                if (args.length != 1 || args[0] == null) return null;
                return new XMLRPC(args[0].toString(), "");
            }
        };

    private static final JSObject.JSFunction soap = new JSObject.JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                if (args.length == 1 && args[0] != null) return new SOAP(args[0].toString(), "", null, null);
                else if (args.length == 2 && args[0] != null && args[1] != null)
                    return new SOAP(args[0].toString(), "", args[1].toString(), null);
                else if (args.length == 3 && args[0] != null && args[1] != null && args[2] != null)
                    return new SOAP(args[0].toString(), "", args[1].toString(), args[2].toString());
                else return null;
            }
        };

    private static final JSObject.JSFunction textwidth = new JSObject.JSFunction() {
            public Object call(Context cx, Scriptable scope, Scriptable thisObj, Object[] args) throws JavaScriptException {
                if (args.length < 1 || args.length > 2) return null;
                if (args[0] == null || (args.length == 2 && args[1] == null)) return null;
                String font = args.length == 1 ? Platform.getDefaultFont() : args[0].toString();
                String text = args.length == 1 ? args[0].toString() : args[1].toString();
                XWF xwf = XWF.getXWF(font);
                if (xwf == null) return new Integer(Platform.stringWidth(font, text));
                else return new Integer(xwf.stringWidth(text));
            }
        };


    private static final JSObject.JSFunction textheight = new JSObject.JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                if (args.length > 1) return null;
                if (args.length == 1 && args[0] == null) return null;
                String font = args.length == 0 || args[0] == null ? Platform.getDefaultFont() : args[0].toString();
                XWF xwf = XWF.getXWF(font);
                if (xwf == null) return new Integer(Platform.getMaxAscent(font) + Platform.getMaxDescent(font));
                else return new Integer(xwf.getMaxAscent() + xwf.getMaxDescent());
            }
        };

    private static final JSObject.JSFunction newBox = new JSObject.JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {

                if (args.length > 0)
                    if (Log.on) Log.log(XWT.class, "DEPRECATED: xwt.newBox() with multiple arguments is deprecated; use xwt.newBox().apply() " +
                                        Context.enter().interpreterSourceFile + ":" + Context.enter().interpreterLine);

                Function callback = null;
                for(int i=1; i<args.length; i++)
                    if (args[i] instanceof Function && callback == null)
                        callback = (Function)args[i];

                Box ret = new Box(args.length == 0 || args[0] == null ? "box" : args[0].toString(), Template.defaultImportList, callback);

                for(int i=1; i<args.length; i++)
                    if (args[i] instanceof Box)
                        ret.put(ret.numChildren(), null, (Box)args[i]);
                for(int i=1; i<args.length; i++)
                    if (args[i] instanceof Scriptable && !(args[i] instanceof Box) && !(args[i] instanceof Function)) {
                        Scriptable s = (Scriptable)args[i];
                        Object[] keys = s.getIds();

                        // FIXME: need to ensure that this is putGlobally(), but still run traps...
                        for(int j=0; j<keys.length; j++) ret.put(keys[j].toString(), null, s.get(keys[j].toString(), s));
                    }
                return ret;
            }
        };

    private static final JSObject.JSFunction sleep = new JSObject.JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                if (args != null && (args.length != 1 || args[0] == null)) return null;
                int i = args == null ? 0 : SpecialBoxProperty.stoi(args[0].toString());

                Thread thread = Thread.currentThread();
                if (!(thread instanceof ThreadMessage)) {
                    if (Log.on) Log.log(this, "cannot sleep() or yield() in the foreground thread");
                    return null;
                }
                ThreadMessage mythread = (ThreadMessage)thread;
                mythread.done.release();

                if (i > 0) try { Thread.sleep(i); } catch (Exception e) { }
                
                MessageQueue.add(mythread);
                mythread.go.block();
                return null;
            }
        };

    private static final JSObject.JSFunction openFile = new JSObject.JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                if (args.length != 1) return null;
                String file = Platform.fileDialog(args[0].toString(), false);
                return file == null ? null : new ByteStream(file);
            }
        };

    private static final JSObject.JSFunction saveFile = new JSObject.JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                if (args.length != 2) return null;
                if (!(args[1] instanceof ByteStream)) return null;
                String file = args[0].toString();
                if (safeFiles.get(Platform.isCaseSensitive() ? file : file.toLowerCase()) == null) {
                    file = Platform.fileDialog(file, true);
                    if (file == null) return null;
                    safeFiles.put(Platform.isCaseSensitive() ? file : file.toLowerCase(), new Object());
                }
                try {
                    ((ByteStream)args[1]).writeTo(new FileOutputStream(file));
                    return null;
                } catch (IOException e) {
                    if (Log.on) Log.log(ByteStream.class, "IO Exception while writing a ByteStream to a file");
                    if (Log.on) Log.log(ByteStream.class, e);
                    throw new JavaScriptException("error while writing a ByteStream to a file");
                }
            }
        };

    private static final JSObject.JSFunction saveFileAs = new JSObject.JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                if (args.length != 2) return null;
                if (!(args[1] instanceof ByteStream)) return null;
                String file = args[0].toString();
                file = Platform.fileDialog(file, true);
                if (file == null) return null;
                safeFiles.put(Platform.isCaseSensitive() ? file : file.toLowerCase(), new Object());
                try {
                    ((ByteStream)args[1]).writeTo(new FileOutputStream(file));
                    return null;
                } catch (IOException e) {
                    if (Log.on) Log.log(ByteStream.class, "IO Exception while writing a ByteStream to a file");
                    if (Log.on) Log.log(ByteStream.class, e);
                    throw new JavaScriptException("error while writing a ByteStream to a file");
                }
            }
        };

    private static final JSObject.JSFunction utfEncode = new JSObject.JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                if (args == null || args.length != 1) return null;
                return new ByteStream(args[0].toString().getBytes());
            }
        };

    
    private static final JSObject.JSFunction parseHTML = new JSObject.JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                if (args == null || args.length != 1 || args[0] == null) return null;
                try {
                    if (args[0] instanceof ByteStream) {
                        return HTML.parseReader(new InputStreamReader(((ByteStream)args[0]).getInputStream()));
                    } else {
                        return HTML.parseReader(new StringReader(args[0].toString()));
                    }
                } catch (IOException e) {
                    if (Log.on) Log.log(HTML.class, "IO Exception while parsing HTML");
                    if (Log.on) Log.log(HTML.class, e);
                    throw new JavaScriptException("error while parsing HTML");
                }
            }
        };
    
    private static void recurse(String indent, String name, Object o, Context cx) {
        if (!name.equals("")) name += " : ";

        if (o == null) {
            Log.log(cx.interpreterSourceFile, indent + name + "<null>");

        } else if (o instanceof NativeArray) {
            Log.log(cx.interpreterSourceFile, indent + name + "<array>");
            NativeArray na = (NativeArray)o;
            for(int i=0; i<na.jsGet_length(); i++)
                recurse(indent + "  ", i + "", na.get(i, null), cx);

        } else if (!(o instanceof NativeDate) && (o instanceof JSObject || o instanceof ScriptableObject)) {
            Log.log(cx.interpreterSourceFile, indent + name + "<object>");
            Scriptable s = (Scriptable)o;
            Object[] keys = s.getIds();
            for(int i=0; i<keys.length; i++)
                recurse(indent + "  ", keys[i].toString(),
                        keys[i] instanceof Integer ? s.get(((Integer)keys[i]).intValue(), null) : s.get(keys[i].toString(), null), cx);

        } else {
            Log.log(cx.interpreterSourceFile, indent + name + o);

        }
    }

    static final JSObject.JSFunction recursivePrintObject = new JSObject.JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) {
                if (args == null || args.length != 1) return null;
                recurse("", "", args[0], cx);
                return null;
            }
        };

    private static final JSObject.JSFunction loadArchive = new JSObject.JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                if (!ThreadMessage.suspendThread()) return null;

                try {
                    if (args == null || args.length < 1 || args[0] == null) return null;
                    URL u = new URL(args[0].toString());
                    
                    Function callback = null;
                    if (args.length == 2 && args[1] != null && args[1] instanceof Function) callback = (Function)args[1];
                    
                    if (!u.getFile().endsWith(".xwar")) {
                        if (Log.on) Log.log(this, "Error: archive names must end with .xwar: " + u.getFile());
                        throw new JavaScriptException("Error: archive names must end with .xwar: " + u.getFile());
                    }
                    
                    if (u.getProtocol().equals("http")) {
                        HTTP http = new HTTP(u.toString());
                        if (Main.originAddr == null) {
                            try {
                                Main.originAddr = InetAddress.getByName(u.getHost());
                            } catch (UnknownHostException e) {
                                if (Log.on) Log.log(this, "couldn't resolve " + u.getHost() + "; proceeding without permissions");
                                Main.originAddr = InetAddress.getByName("0.0.0.0");
                            }
                        } else {
                            Main.originAddr = InetAddress.getByName("0.0.0.0");
                        }
                        HTTP.HTTPInputStream in = http.GET();
                        Resources.loadArchive(in, in.getContentLength(), callback);

                    } else if (u.getProtocol().equals("file")) {
                        if (Main.originAddr != null) {
                            if (Log.on) Log.log(this, "scripts downloaded from the network may not load xwars from the local filesystem");
                            throw new JavaScriptException("scripts downloaded from the network may not load xwars from the local filesystem");
                        }
                        Resources.loadArchive(new FileInputStream(u.getFile()), (int)new File(u.getFile()).length(), callback);
                        
                    } else {
                        if (Log.on) Log.log(this, "unknown protocol \"" + u.getProtocol() + "\"");
                        throw new JavaScriptException("unknown protocol \"" + u.getProtocol() + "\"");
                    }
                    
                } catch (MalformedURLException me) {
                    if (Log.on) Log.log(this, "Malformed URL: " + args[0]);
                    if (Log.on) Log.log(this, me);
                    throw new JavaScriptException(me.toString());
                    
                } catch (IOException ioe) {
                    if (Log.on) Log.log(this, "IOException while loading archive:");
                    if (Log.on) Log.log(this, ioe);
                    throw new JavaScriptException(ioe.toString());

                } finally {
                    ThreadMessage.resumeThread();

                }
                return null;
            }
        };

    private static final JSObject.JSFunction prefetchImage = new JSObject.JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                if (args == null || args.length < 1 || args[0] == null) return null;
                Box.getImage(args[0].toString(), args.length > 1 && args[1] instanceof Function ? (Function)args[1] : null);
                return null;
            }
        };

}





